from typing import Any, Sequence
import argparse


class DangrArgparse(argparse.ArgumentParser):
    """
    Custom argument parser for the 'dangr' project.

    Provides the cli for all the arguments needed for the dangr_rt and
    allows the user to add other arguments if needed.

    In the resulting parserd arguments a `config` dictionary is
    accesible. That same dict can be recieved by the DangrAnalysis.
    """
    def __init__(self, description: str) -> None:
        super().__init__(description=description)
        self._config: dict[str, Any] = {}

        self._add_dangr_argument(
            "max_depth",
            "-d",
            "--max-depth",
            type=int,
            default=None,
            help="Maximum depth for backward execution."
        )

        self._add_dangr_argument(
            "timeout",
            "-t",
            "--timeout",
            type=int,
            default=None,
            help="Timeout for dangr simulation step."
        )

    def _add_dangr_argument(self, config_key: str, *args: Any, **kwargs: Any) -> None:
        self._config[config_key] = str(kwargs.get('default'))
        super().add_argument(*args, **kwargs)

    def dangr_parse_args(
        self,
        args: Sequence[str] | None,
        namespace: argparse.Namespace | None = None
    ) -> argparse.Namespace:
        """
        Parses arguments, updates the config dictionary only with fixed argument values,
        and attaches config as an attribute of the returned namespace.
        """
        parsed_args = self.parse_args(args, namespace)

        for key in self._config:
            if hasattr(parsed_args, key):
                self._config[key] = getattr(parsed_args, key)

        setattr(parsed_args, "config", self._config)
        return parsed_args



from typing import Final
from itertools import product

from networkx import DiGraph # type: ignore [import-untyped]
from networkx.algorithms import has_path # type: ignore [import-untyped]

import angr

from dangr_rt.dangr_types import Address
from dangr_rt.variables import Variable
from dangr_rt.variable_factory import VariableFactory

class DependencyAnalyzer:
    """
    A class for analyzing dependencies between variables in a binary program using a
    Dependency Dependency Graph (DDG).
    """
    CALL_DEPTH_DEFAULT: Final = 1

    def __init__(
        self, project: angr.Project,
        variable_factory: VariableFactory,
        call_depth: int | None = None
    ):
        self.project = project
        self.ddg: DiGraph | None = None
        self.call_depth = call_depth or self.CALL_DEPTH_DEFAULT
        self.variable_factory = variable_factory

    def create_dependency_graph(self, start_address: Address) -> None:
        """
        Create a Dependency Dependency Graph (DDG) starting from the given address.

        Args:
            start_address (Address): The starting address for the DDG creation.
        """
        cfg = self.project.analyses.CFGEmulated(
            keep_state=True,
            starts=[start_address],
            call_depth=self.call_depth,
            state_add_options=angr.sim_options.refs | {angr.sim_options.NO_CROSS_INSN_OPT},
        )

        self.ddg = self.project.analyses.DDG(cfg=cfg, start=start_address)

    def _find_reference_nodes(self, addr: Address) -> list[angr.code_location.CodeLocation]:
        return [node for node
            in self.ddg.graph.nodes() # type: ignore [union-attr] # already chequed where its called
            if node.ins_addr == addr]

    def check_dependency(self, source: Variable, target: Variable) -> bool:
        """
        Check if `source` affects the value of `target`

        Args:
            source (Variable): The source variable to check for dependencies.
            target (Variable): The target variable to check for dependencies.

        Returns:
            bool: True if a dependency path is found from source to target, False otherwise.

        Raises:
            ValueError: If the dependency graph has not been created.
        """
        if not self.ddg:
            raise ValueError("Dependency graph is None. Call create_dependency_graph() first.")

        return any(
            has_path(self.ddg.graph, src_node, trg_node)
            for src_node, trg_node in product(
                self._find_reference_nodes(source.ref_addr),
                self._find_reference_nodes(target.ref_addr)
            )
        )



from abc import ABC, abstractmethod
from typing import override, Final, Callable, Any
from dataclasses import dataclass
from copy import deepcopy
import angr

from dangr_rt.variables import Variable
from dangr_rt.dangr_types import Address, CFGNode
ConcreteState = dict['Variable', int]

EXTERNAL_ADDR_SPACE_BASE: Final = 0x500000
ENDBR64_MNEMONIC: Final = 'endbr64'


class Simulator(ABC):
    """
    A manager for symbolic execution simulations using angr. 

    This class is designed to handle symbolic execution.

    Attributes:
        project (angr.Project): The angr project to simulate.
        init_addr (Address): The initial address where the simulation starts.
    """
    def __init__(self, project: angr.Project) -> None:
        self.project = project
        self.initial_values: None | ConcreteState = None

    def _initialize_state(
        self,
        start: Address,
        add_options: set[str] | None = None
    ) -> angr.SimState:
        """"
        Sets the initial values from concrete_state in the given state
        Modifies state
        """
        state: angr.SimState = self.project.factory.blank_state( # type: ignore [no-untyped-call]
            addr=start,
            add_options=add_options
        )

        if self.initial_values is None:
            return state

        for var, value in self.initial_values.items():
            var.set_ref_states([state])
            var.set_value(value)

        return state

    def set_initial_values(self, initial_values: ConcreteState) -> None:
        """
        Set values for the initial state of the simulation
        """
        self.initial_values = initial_values

    @abstractmethod
    def simulate(self) -> list[angr.SimState]:
        """
        Excecutes simbolically.

        Arguments:
            concrete_state (ConcreteState): the values to initialize the simulation

        Returns
            list[angr.SimState]: the list of states on which the target was found

        """


class ForwardSimulation(Simulator):
    """
    Simulate until reaching a target
    """
    def __init__(self, project: angr.Project, init_addr: Address, target: Address) -> None:
        super().__init__(project)
        self.init_addr = init_addr
        self.target = target

    @override
    def simulate(self) -> list[angr.SimState]:
        found_states = []

        initial = self._initialize_state(self.init_addr)
        simulation = self.project.factory.simulation_manager(initial)

        initial.inspect.b( # type: ignore [no-untyped-call]
            'instruction', when=angr.BP_AFTER, instruction=self.target,
            action=lambda state: found_states.append(deepcopy(state))
        )

        while simulation.active:
            simulation.step() # type: ignore [no-untyped-call]

        return found_states

class StepSimulation(Simulator):
    """
    Simulates a chunk of a binary from the init_addr until the first address is met.
    The simulation can be resumed from the previous state.
    """

    def __init__(
        self,
        project: angr.Project,
        init_addr: Address,
        timeout: int | None = None
    ) -> None:

        super().__init__(project)
        self.init_addr = init_addr

        self.target: Address | None = None
        self.previous_states: list[angr.SimState] | angr.SimState | None = None
        self.timeout = timeout

    def set_step_target(self, target: Address) -> None:
        self.target = target

    def simulate(self) -> list[angr.SimState]:
        if self.previous_states is None:
            initial = self._initialize_state(self.init_addr)
            self.previous_states = initial

        simulation = self.project.factory.simulation_manager(self.previous_states)
        timeout_tech = angr.exploration_techniques.Timeout(self.timeout) # type: ignore [no-untyped-call]
        simulation.use_technique(timeout_tech) # type: ignore [no-untyped-call]
        simulation.explore(find=self.target) # type: ignore [no-untyped-call]
        self.previous_states = simulation.found

        return simulation.found

@dataclass
class RecursiveCtx:
    current_depth: int
    backup_state: angr.SimState | None
    path: list[CFGNode]

class BackwardSimulation(Simulator):
    """
    Simualte backwards until variables are concrete
    """
    def __init__( # pylint: disable=too-many-arguments
        self, project: angr.Project, *, target: Address,
        cfg: angr.analyses.CFGFast, variables: list[Variable],
        max_depth: int | None = None
    ) -> None:

        super().__init__(project)
        self.target = target
        self.cfg = cfg
        self.variables = variables
        self.states_found: list[angr.SimState] = []
        self.max_depth: Final = max_depth or 1

    @override
    def simulate(self) -> list[angr.SimState]:
        target_node = self.cfg.model.get_any_node(self.target)

        if target_node is None:
            raise ValueError("Target node not found")

        rec_ctx = RecursiveCtx(0, None, [target_node])
        self._rec_simulate(target_node, rec_ctx)
        return self.states_found

    def _node_addr(self, node: CFGNode) -> Address:
        if not isinstance(node.addr, int):
            raise ValueError(f"Unsupported node {node}")
        return node.addr

    def _rec_simulate(
        self,
        target_node: CFGNode,
        rec_ctx: RecursiveCtx
    ) -> None:
        """
        Simulate the execution of the program until reaching the `function_node`, the simulation
        is done on the `path` and starts on the last node of the path.
        
        It will recursively look for the path that makes the `registers` concrete and when found
        it returns the concrete values of the registers in that path.
        """
        initial_node = rec_ctx.path[-1]
        state = self._simulate_slice(self._node_addr(initial_node), target_node, rec_ctx.path)

        if not state:
            self._handle_no_found_state(rec_ctx)
            return

        self._set_state_to_vars(state)

        if self._rec_simulation_stop(rec_ctx):
            self.states_found.append(state)
            return

        for pred in list(self.cfg.model.get_predecessors(initial_node)):
            new_rec_ctx = RecursiveCtx(rec_ctx.current_depth + 1, state, rec_ctx.path + [pred])
            self._rec_simulate(target_node, new_rec_ctx)

    def _handle_no_found_state(self, rec_ctx: RecursiveCtx) -> None:
        if rec_ctx.backup_state:
            self.states_found.append(rec_ctx.backup_state)

    def _set_state_to_vars(self, state: angr.SimState) -> None:
        for var in self.variables:
            var.set_ref_states([state])

    def _rec_simulation_stop(self, rec_ctx: RecursiveCtx) -> bool:
        happy_stop =  all(var.is_concrete() for var in self.variables)
        forced_stop = rec_ctx.current_depth >= self.max_depth
        return happy_stop or forced_stop

    def _simulate_slice(
        self,
        start: Address,
        target_node: CFGNode,
        pred: list[CFGNode],
    ) -> angr.SimState | None:

        initial = self._initialize_state(start)
        simgr = self.project.factory.simulation_manager(initial)
        state_found = self._get_finding(simgr, self._node_addr(target_node))

        while simgr.active and not state_found:
            self._remove_states(simgr.active, pred)
            simgr.step() # type: ignore [no-untyped-call]
            state_found = self._get_finding(simgr, self._node_addr(target_node))

        return state_found

    def _get_finding(self, simgr: angr.SimulationManager, target: Address) -> angr.SimState | None:
        return next((state for state in simgr.active if state.addr == target), None)

    def _remove_states(self, active_states: list[angr.SimState], pred: list[CFGNode]) -> None:
        for state in active_states:
            if self._remove_condition(state, pred):
                active_states.remove(state)

    def _remove_condition(self, state: angr.SimState, pred: list[CFGNode]) -> bool:
        already_visited = state.addr in state.history.bbl_addrs
        is_external_block =  state.addr >= EXTERNAL_ADDR_SPACE_BASE
        is_in_slice = state.addr in [p.addr for p in pred]
        return not is_in_slice or is_external_block or already_visited

class HookSimulation(Simulator):
    """
    Simulate until reaching a target
    """
    def __init__(
        self,
        project: angr.Project,
        init_addr: Address,
        stop: Callable[[list[angr.SimState]], bool],
        **inspect_kwargs: Any
    ) -> None:

        super().__init__(project)
        self.init_addr = init_addr
        self.stop = stop
        self.inspect_kwargs = inspect_kwargs

    @override
    def simulate(self) -> list[angr.SimState]:
        initial = self._initialize_state(self.init_addr)
        simulation = self.project.factory.simulation_manager(initial)
        initial.inspect.b(**self.inspect_kwargs) # type: ignore [no-untyped-call]

        while simulation.active and not self.stop(simulation.active):
            simulation.step() # type: ignore [no-untyped-call]

        return simulation.active


class BackwardSliceSimulation(Simulator):
    """
    TODO: This only works when init_addr and target are the begggining of a block
    """
    def __init__(self, project: angr.Project, init_addr: Address, target: Address) -> None:
        super().__init__(project)
        self.init_addr = init_addr
        self.target = target

    def _annotated_cfg(
        self,
        cfg: angr.analyses.CFGEmulated,
        target_node: CFGNode
    ) -> angr.analyses.CFGEmulated:

        bs = self.project.analyses.BackwardSlice(
            cfg, None, None,
            targets=[(target_node, -1)],
            control_flow_slice=True
        )
        acfg: angr.analyses.CFGEmulated = bs.annotated_cfg() # type: ignore [no-untyped-call]
        return acfg

    def _target_node(self, cfg: angr.analyses.CFGEmulated) -> CFGNode:
        cfg_nodes: list[CFGNode] = cfg.nodes()
        for node in cfg_nodes:
            if node.size and node.addr <= self.target < node.addr + node.size:
                return node
        raise ValueError("Target node was not found")

    @override
    def simulate(self) -> list[angr.SimState]:
        initial = self._initialize_state(self.init_addr, {angr.options.LAZY_SOLVES})
        simgr = self.project.factory.simgr(initial) # type: ignore [no-untyped-call]

        cfg = self.project.analyses.CFGEmulated(keep_state=True, starts=[self.init_addr])
        target_node = self._target_node(cfg)
        slicecutor = angr.exploration_techniques.Slicecutor(self._annotated_cfg(cfg, target_node))
        simgr.use_technique(slicecutor)

        found_states: list[angr.SimState] = []
        initial.inspect.b( # type: ignore [no-untyped-call]
            'instruction', when=angr.BP_AFTER, instruction=target_node.addr,
            action=lambda state: found_states.append(deepcopy(state))
        )

        while simgr.active:
            simgr.step()

        return found_states



from typing import override, Sequence, Any
from abc import ABC, abstractmethod
from itertools import product
import claripy
from dangr_rt.dangr_types import AngrExpr, Address, BYTE_SIZE, AngrBool, AngrArith
from dangr_rt.variables import Variable

class  Expression[TypeResult](ABC):
    """
    Represents an expression. It can be boolean or arithmetic
    """
    @abstractmethod
    def get_expr(self) -> Sequence[TypeResult]:
        """
        Returns a list of the possible angr representations of this expression
        """

    @property
    @abstractmethod
    def ref_addr(self) -> Address | None:
        """
        Returns the reference address in the binary of the expression.
        It is calculated as the address of the atom with the last reference or
        None if the expression is all based on constants that do not appear in the binary 
        """

    @staticmethod
    def _unique(list_w_repeated_elems: list[Any]) -> list[Any]:
        return list(set(list_w_repeated_elems))

    @abstractmethod
    def _to_str(self) -> str:
        pass

    def __repr__(self) -> str:
        return self._to_str()

class Binary[TypeLeft, TypeRight, TypeResult](Expression[TypeResult]):
    """
    Abstract class that represents a binary expression
    """
    def __init__(
        self,
        lhs: Expression[TypeLeft] | Variable | int | bool,
        rhs: Expression[TypeRight] | Variable | int | bool,
        op: str
    ):
        self.op: str = op
        self.lhs = lhs
        self.rhs = rhs

    @override
    @property
    def ref_addr(self) -> Address | None:
        lhs_addr: int | None = getattr(self.lhs, 'ref_addr', None)
        rhs_addr: int | None = getattr(self.rhs, 'ref_addr', None)

        if lhs_addr is None:
            return rhs_addr
        if rhs_addr is None:
            return lhs_addr

        return max(lhs_addr, rhs_addr)

    @override
    def _to_str(self) -> str:
        return f'[{self.lhs!r} {self.op} {self.rhs!r}]'

    def _operands_product(self) -> list[tuple[TypeLeft, TypeRight]]:
        lhs_exprs = getattr(self.lhs, 'get_expr', lambda: [self.lhs])()
        rhs_exprs = getattr(self.rhs, 'get_expr', lambda: [self.rhs])()

        return list(product(lhs_exprs, rhs_exprs))

class Eq(Binary[AngrExpr, AngrExpr, AngrBool]):
    """
    Eq(lhs, rhs) represents the constraint (lhs == rhs)
    """
    def __init__(
        self,
        lhs: Expression[AngrExpr] | Variable | int | bool,
        rhs: Expression[AngrExpr] | Variable | int | bool
    ):
        super().__init__(lhs, rhs, op='==')

    @override
    def get_expr(self) -> Sequence[AngrBool]:
        return self._unique([lh == rh for lh, rh in self._operands_product()])

class And(Binary[AngrBool, AngrBool, AngrBool]):
    """
    And(lhs, rhs) represents the constraint (lhs & rhs)
    """
    def __init__(self, lhs: Expression[AngrBool] | bool, rhs: Expression[AngrBool] | bool):
        super().__init__(lhs, rhs, op='&')

    @override
    def get_expr(self) -> Sequence[AngrBool]:
        return self._unique([claripy.And(lh, rh) for lh, rh in self._operands_product()])

class Or(Binary[AngrBool, AngrBool, AngrBool]):
    """
    Or(lhs, rhs) represents the constraint (lhs | rhs)
    """
    def __init__(self, lhs: Expression[AngrBool] | bool, rhs: Expression[AngrBool] | bool):
        super().__init__(lhs, rhs, op='|')

    @override
    def get_expr(self) -> Sequence[AngrBool]:
        return self._unique([claripy.Or(lh, rh) for lh, rh in self._operands_product()])

class Not(Expression[AngrBool]):
    """
    Not(operand) represents the constraint ~operand
    """
    def __init__(self, operand: Expression[AngrBool] | bool) -> None:
        self.operand = operand

    @override
    def get_expr(self) -> Sequence[AngrBool]:
        if isinstance(self.operand, bool):
            return [claripy.BoolV(self.operand)]
        exprs = getattr(self.operand, 'get_expr', lambda: claripy.BoolV(self.operand))()
        return self._unique([claripy.Not(op_expr) for op_expr in exprs])

    @override
    @property
    def ref_addr(self) -> Address | None:
        return getattr(self.operand, 'ref_addr', None)

    @override
    def _to_str(self) -> str:
        return f'~{self.operand!r}'

class Add(Binary[AngrArith, AngrArith, AngrArith]):
    """
    Add(lhs, rhs) represents the operation lhs + rhs
    """
    def __init__(
        self,
        lhs: Expression[AngrArith] | Variable | int,
        rhs: Expression[AngrArith] | Variable | int
    ) -> None:
        super().__init__(lhs, rhs, op='+')

    @override
    def get_expr(self) -> Sequence[AngrArith]:
        return self._unique([lh + rh for lh, rh in self._operands_product()])

class Mul(Binary[AngrArith, AngrArith, AngrArith]):
    """
    Mul(lhs, rhs) represents the operation lhs + rhs
    """
    def __init__(
        self,
        lhs: Expression[AngrArith] | Variable | int,
        rhs: Expression[AngrArith] | Variable | int
    ) -> None:
        super().__init__(lhs, rhs, op='*')

    @override
    def get_expr(self) -> Sequence[AngrArith]:
        return self._unique([lh * rh for lh, rh in self._operands_product()])

class Sub(Binary[AngrArith, AngrArith, AngrArith]):
    """
    Sub(lhs, rhs) represents the operation lhs - rhs
    """
    def __init__(
        self,
        lhs: Expression[AngrArith] | Variable | int,
        rhs: Expression[AngrArith] | Variable | int
    ) -> None:
        super().__init__(lhs, rhs, op='-')

    @override
    def get_expr(self) -> Sequence[AngrArith]:
        return self._unique([lh - rh for lh, rh in self._operands_product()])

class Div(Binary[AngrArith, AngrArith, AngrArith]):
    """
    Div(lhs, rhs) represents the operation lhs // rhs (integer division)
    """
    def __init__(
        self,
        lhs: Expression[AngrArith] | Variable | int,
        rhs: Expression[AngrArith] | Variable | int
    ) -> None:
        super().__init__(lhs, rhs, op='//')

    @override
    def get_expr(self) -> Sequence[AngrArith]:
        return self._unique([lh // rh for lh, rh in self._operands_product()])

class IsMax(Expression[AngrBool]):
    """
    IsMax(operand) represents the constraint operand == <operand max value>
    """
    def __init__(self, operand: Expression[AngrArith]):
        self.operand = operand

    def _size(self, angr_exp: AngrArith) -> int:
        if isinstance(angr_exp, (int, bool)):
            raise ValueError("Can't calculate the max value of an int or bool", angr_exp)

        return angr_exp.size()

    def _max_value(self, angr_exp: AngrArith) -> int:
        max_value: int = 2**self._size(angr_exp)*BYTE_SIZE
        return max_value

    @override
    def get_expr(self) -> Sequence[AngrBool]:
        return self._unique([
            exp == self._max_value(exp) for exp in self.operand.get_expr()
        ])

    @override
    @property
    def ref_addr(self) -> Address | None:
        return self.operand.ref_addr

    @override
    def _to_str(self) -> str:
        return f'IsMax({self.operand!r})'



"""
This is a mocked version of JASM
It will be modified and correctly implemented when
JASM has support for some functionalities I need
which is already in progres but not finished.
"""

from typing import Any, cast
import json
from dataclasses import dataclass
from dangr_rt.dangr_types import Address

def load_json(file_path: str) -> dict[str, Any] | None:
    try:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = json.load(file)
        return cast(dict[str, Any], data)
    except json.JSONDecodeError as e:
        print("Error decoding JSON:", e)
        return None
    except FileNotFoundError:
        print("File not found:", file_path)
        return None


@dataclass
class VariableMatch:
    name: str
    value: str
    addr: int

@dataclass
class AddressMatch:
    name: str
    value: int

# jasm returns a set of JasmMatch's
class JasmMatch:
    def __init__(
        self,
        match: dict[int, str],
        variables: list[VariableMatch],
        address_captures: list[AddressMatch]
    ) -> None:
        self._match = match
        self._variables = variables
        self._address_captures = address_captures

    @property
    def variables(self) -> list[VariableMatch]:
        """
        Returns a dict with all the variables matched
        Including
        - The variable name
        - The register/literal matched
        - The address capture in the instruction if it exists
        """
        return self._variables

    @property
    def address_captures(self) -> list[AddressMatch]:
        """
        Returns a list with all the address captures.
        The keys are the anchore's names and the value is the match
        """
        return self._address_captures

    def addrmatch_from_name(self, capture_name: str) -> AddressMatch:
        for addr_capt in self._address_captures:
            if addr_capt.name == capture_name:
                return addr_capt
        raise ValueError("Invalid capture name")

    def varmatch_from_name(self, capture_name: str) -> VariableMatch:
        for varmatch in self._variables:
            if varmatch.name == capture_name:
                return varmatch
        raise ValueError("Invalid capture name")

    @property
    def start(self) -> Address:
        return list(self._match.keys())[0]

    @property
    def end(self) -> Address:
        return list(self._match.keys())[-1]

def cast_var_value(var_value: str | int) -> int | str:
    try:
        return int(var_value, 16)
    except Exception:
        return var_value

def _parse_jasm_output(jasm_out: list[dict[str, Any]]) -> list[JasmMatch]:
    all_matches = []
    for j_match in jasm_out:
        out_match = {}
        out_vars = []
        out_addr_capt = []
        for m in j_match['match']:
            addr, instr = m.split("::")
            out_match[int(addr, 16) + 0x40_0000] = instr

        for name_capt, info_capt in j_match['name-captures'].items():
            out_vars.append(
                VariableMatch(
                    name=name_capt,
                    value=cast_var_value(info_capt[0]),
                    addr=int(info_capt[1], 16) + 0x40_0000
                )
            )
        for addr_capt, info_capt in j_match['address-captures'].items():
            out_addr_capt.append(
                AddressMatch(
                    name=addr_capt,
                    value=int(info_capt, 16) + 0x40_0000
                ))

        all_matches.append(
            JasmMatch(
                match=out_match,
                variables=out_vars,
                address_captures=out_addr_capt
        ))

    return all_matches

def _run_jasm(jasm_pattern: str, binary_path: str) -> list[JasmMatch]:
    path_to_mock = "jasm_mock/"
    _ = binary_path
    match jasm_pattern:
        case 'software_breakpoint_pattern':
            path_to_mock += "sw_brk_jasm_out.json"
        case _:
            print("not implemented")

    jasm_match_uparsed = load_json(path_to_mock)
    return _parse_jasm_output(jasm_match_uparsed)


def structural_filter(binary_path: str, jasm_pattern) -> list[JasmMatch]:
    jasm_matches = _run_jasm(jasm_pattern, binary_path)
    return jasm_matches



from collections import namedtuple
import angr

from dangr_rt.variables import Variable
from dangr_rt.simulator import StepSimulation, ConcreteState
from dangr_rt.dangr_types import Address, AngrBool
from dangr_rt.expression import Expression

CheckpointGroup = namedtuple('CheckpointGroup', ['variables', 'constraints'])

class Checkpoints(dict[Address, CheckpointGroup]):

    def add_variable(self, address: Address, variable: Variable) -> None:
        if address not in self:
            self[address] = CheckpointGroup([], [])

        self[address].variables.append(variable)

    def add_constraint(self, address: Address, constraint: Expression[AngrBool]) -> None:
        if address not in self:
            self[address] = CheckpointGroup([], [])

        self[address].constraints.append(constraint)

    def sorted(self) -> 'Checkpoints':
        """
        Return a new Checkpoints object with items sorted by the dictionary keys.
        """
        sorted_checkpoints = Checkpoints(sorted(self.items()))
        return sorted_checkpoints

    def last_address(self) -> Address | None:
        if not self:
            return None

        last_key = next(reversed(self.sorted()), None)
        return last_key


class DangrSimulation:
    def __init__(
        self,
        project: angr.Project,
        init_addr: Address,
        timeout: int | None = None
    ) -> None:

        self.simulator = StepSimulation(project, init_addr, timeout)
        self.variables: list[Variable] = []
        self.constraints: list[Expression[AngrBool]] = []

    def add_variables(self, variables: list[Variable]) -> None:
        self.variables.extend(variables)

    def add_constraints(self, constraints: list[Expression[AngrBool]]) -> None:
        self.constraints.extend(constraints)

    def simulate(
        self,
        target: Address,
        init_states: list[ConcreteState] | None = None
    ) -> list[angr.SimState]:
        """
        Symbolic execute the current function until the target is found
        """
        checkpoints = self._create_checkpoints(target)

        for addr, action_elem in checkpoints.items():
            found_states: list[angr.SimState] = []
            self.simulator.set_step_target(target=addr)

            if not init_states:
                found_states.extend(self.simulator.simulate())
            else:
                for init_state in init_states:
                    self.simulator.set_initial_values(init_state)
                    found_states.extend(self.simulator.simulate())

            self._set_states_to_vars(action_elem.variables, found_states)
            self._add_constraints_to_states(action_elem.constraints, found_states)

        return found_states

    def _set_states_to_vars(self, variables: list[Variable], states: list[angr.SimState]) -> None:
        for var in variables:
            var.set_ref_states(states)

    def _add_constraints_to_states(
        self,
        constraints: list[Expression[AngrBool]],
        states: list[angr.SimState]
    ) -> None:

        for constraint in constraints:
            for expr in constraint.get_expr():
                for state in states:
                    state.solver.add(expr)


    def _create_checkpoints(self, target: Address) -> Checkpoints:
        checkpoints = Checkpoints()
        self._create_var_checkpoints(checkpoints)
        self._create_constr_checkpoints(checkpoints, target)
        self._add_target_checkpoint(checkpoints, target)
        return checkpoints.sorted()

    def _create_var_checkpoints(self, checkpoints: Checkpoints) -> None:
        for variable in self.variables:
            checkpoints.add_variable(variable.ref_addr, variable)

    def _create_constr_checkpoints(self, checkpoints: Checkpoints, default_addr: Address) -> None:
        for constraint in self.constraints:
            checkpoints.add_constraint(constraint.ref_addr or default_addr, constraint)

    def _add_target_checkpoint(self, checkpoints: Checkpoints, target: Address) -> None:
        if checkpoints.last_address() is None or\
           checkpoints.last_address() < target: # type: ignore [operator]
            checkpoints[target] = CheckpointGroup([], [])



from dataclasses import dataclass
import angr
import claripy

Address = int
Path = str
CFGNode = angr.knowledge_plugins.cfg.cfg_node.CFGNode
RegOffset = int

AngrArith = claripy.ast.bv.BV | int
AngrBool = claripy.ast.bool.Bool | bool

AngrExpr = AngrArith | AngrBool
BYTE_SIZE = 8

@dataclass
class Argument:
    """
    A data class representing an argument in a function call.
    """
    idx: int
    call_address: Address
    size: int



from typing import Final, Any, Sequence
import angr
from dangr_rt.jasm_findings import JasmMatch
from dangr_rt.dangr_types import Address, Path, AngrBool
from dangr_rt.variables import Variable, Register
from dangr_rt.variable_factory import VariableFactory
from dangr_rt.expression import Expression
from dangr_rt.simulator import ConcreteState
from dangr_rt.arguments_analyzer import ArgumentsAnalyzer
from dangr_rt.dependency_analyzer import DependencyAnalyzer
from dangr_rt.dangr_simulation import DangrSimulation

class DangrAnalysis: # pylint: disable=too-many-instance-attributes
    """
    Class that provides all the interface necesary for the analysis
    """

    def __init__(self, binary_path: Path, config: dict[str, Any]) -> None:
        """
        Here we set all the attributes that are independent from the structural finding
        """
        # general config
        self.project: Final = angr.Project(binary_path, load_options={"auto_load_libs": False})
        self.cfg: Final = self.project.analyses.CFGFast()
        self.config = config
        self.current_function: Address | None = None
        self.jasm_match: JasmMatch | None = None

        # helper modules init
        self.simulator: DangrSimulation | None = None
        self.variable_factory = VariableFactory(self.project)
        self.dependency_analyzer = DependencyAnalyzer(self.project, self.variable_factory)
        self.arguments_analyzer = ArgumentsAnalyzer(self.project,
                                                    self.cfg,
                                                    self.config.get('max_depth', None))

    def _jasm_match_set(self) -> None:
        """
        Checks if self.simulator, self.jasm_match, and self.current_function
        are not None. Raises ValueError if any of them is None, indicating that
        set_finding should be called first.
        """
        if self.simulator is None or self.jasm_match is None or self.current_function is None:
            raise ValueError("Analysis not properly initialized. Call `set_finding()` first.")

    def add_variables(self, variables: list[Variable]) -> None:
        self._jasm_match_set()
        self.simulator.add_variables(variables) # type: ignore [union-attr]

    def set_finding(self, jasm_match: JasmMatch) -> None:
        """
        Sets the structural finding and updates the current function.

        Args:
            jasm_match (JasmMatch): The new structural finding to set.
        """
        self.jasm_match = jasm_match

        # Restart analysis
        self.current_function = self._find_function()
        self.simulator = DangrSimulation(
            project=self.project,
            init_addr=self.current_function,
            timeout=self.config.get('timeout', None)
        )
        self.dependency_analyzer.create_dependency_graph(self.current_function)

    def get_variable_factory(self) -> VariableFactory:
        return self.variable_factory

    def _find_function(self) -> Address:
        """
        Gets the address of the function that contains the structural pattern matched.

        Raises:
            ValueError: If no single function contains the matched address range, 
            typically caused by the jasm pattern spanning multiple functions.
        """
        if self.jasm_match is None:
            raise ValueError('Structural finding not set')

        for fn in self.cfg.kb.functions.values():
            if self._finding_in_func(fn):
                return Address(fn.addr)

        raise ValueError('Function not found for target address: '
        f'start at {hex(self.jasm_match.start)}, end at {hex(self.jasm_match.end)}')

    def _finding_in_func(self, fn: angr.knowledge_plugins.functions.function.Function) -> bool:
        if self.jasm_match is None:
            raise ValueError('Structural finding not set')

        return bool(
            (fn.addr <= self.jasm_match.start) and\
            (self.jasm_match.end <= fn.addr + fn.size))

    def get_fn_args(self) -> Sequence[Register]:
        self._jasm_match_set()
        return self.arguments_analyzer.get_fn_args(self.current_function) # type: ignore [arg-type]

    def concretize_fn_args(self) -> list[ConcreteState]:
        """
        Returns a list with the concrete possible values of the arguments used
        in the function being analyzed

        Returns:
            list[ConcreteState]: all the possible combinations of the arguments values
        """
        self._jasm_match_set()

        return self.arguments_analyzer.solve_arguments(
            # already checked in _jasm_match_set() ↓
            self.current_function, # type: ignore [arg-type]
            self.get_fn_args()
        )

    def simulate(
        self,
        target: Address,
        init_states: list[ConcreteState] | None = None
    ) -> list[angr.SimState]:
        self._jasm_match_set()
        return self.simulator.simulate(target, init_states) # type: ignore [union-attr]

    def add_constraint(self, constraint: Expression[AngrBool]) -> None:
        """
        Adds a constraints to the analysis
        """
        self._jasm_match_set()
        self.simulator.add_constraints([constraint]) # type: ignore [union-attr]

    def satisfiable(self, states: list[angr.SimState]) -> bool:
        """
        Returns True if all the constraints can be satisfied at the same time
        in any of the states given.
        """
        for state in states:
            state.solver.reload_solver() # type: ignore [no-untyped-call]
            if state.solver.satisfiable():
                return True
        return False

    def depends(self, source: Variable, target: Variable) -> bool:
        """
        Calculates dependencies of a given variable
        """
        return self.dependency_analyzer.check_dependency(source, target)



from typing import Final, Sequence, cast
from collections import namedtuple
import angr
from dangr_rt.variables import Register
from dangr_rt.simulator import BackwardSimulation, HookSimulation, ConcreteState
from dangr_rt.dangr_types import Address, RegOffset

RegInfo = namedtuple('RegInfo', ['size', 'addr'])
SimRegArg = angr.calling_conventions.SimRegArg

class ArgumentsAnalyzer:
    """
    This class is responsible for obtaining the concrete values of registers in a given project.
    Implements the method `solve_registers`.
    """
    def __init__(
        self,
        project: angr.Project,
        cfg: angr.analyses.CFGFast,
        max_depth: int | None = None
    ) -> None:
        self.project: Final = project
        self.cfg: Final = cfg
        self.max_depth = max_depth
        self.args_info: dict[RegOffset, RegInfo | None] = {}

    def get_fn_args(self, fn_addr: Address) -> Sequence[Register]:
        """
        Returns the arguments of the function from address `fn_addr`
        """
        args = self._get_args(fn_addr)
        self._init_args_info(args)

        HookSimulation(
            project=self.project,
            init_addr=fn_addr,
            stop=lambda sts: all(self.args_info.values()),
            event_type='reg_read',
            action=self._record_reg_read,
            when=angr.BP_BEFORE,
        ).simulate()

        return self._create_arg_regs()

    def _create_arg_regs(self) -> Sequence[Register]:
        found_args = []
        for offset, found_info in self.args_info.items():
            if found_info is None:
                raise ValueError("Argument couldn't be found")

            reg_name = self.project.arch.register_size_names[offset, found_info.size]
            found_args.append(Register(self.project, reg_name, found_info.addr))
        return found_args

    def _get_args(self, fn_addr: Address) -> list[SimRegArg]:
        func = self.cfg.functions.get(fn_addr)
        self.project.analyses.VariableRecoveryFast(func)
        cca = self.project.analyses.CallingConvention(func, self.cfg.model, analyze_callsites=True)

        if cca.cc is None:
            raise ValueError("Unsupported calling convention")

        args = cca.cc.arg_locs(cca.prototype)

        for arg in args:
            if not isinstance(arg, SimRegArg):
                raise TypeError(f"Unsupported argument {arg}")

        return cast(list[SimRegArg], args)

    def _init_args_info(self, args: list[SimRegArg]) -> None:
        for arg in args:
            offset: int = arg.check_offset(self.project.arch) # type: ignore [no-untyped-call]
            self.args_info[offset] = None

    def _record_reg_read(self, state: angr.SimState) -> None:
        """Record the instruction address of the first read of the register."""
        addr = self._record_addr(state)
        offset = self._record_offset(state)

        if offset in self.args_info and self.args_info[offset] is None:
            block = state.block(addr) # type: ignore [no-untyped-call]
            size = block.capstone.insns[0].insn.operands[0].size
            self.args_info[offset] = RegInfo(size, addr)

    def _record_addr(self, state: angr.SimState) -> Address:
        if hasattr(state.inspect, 'instruction'):
            return Address(state.inspect.instruction)
        raise ValueError("Instruction address couldn't be found")

    def _record_offset(self, state: angr.SimState) -> RegOffset:
        if hasattr(state.inspect, 'reg_read_offset'):
            return state.solver.eval(state.inspect.reg_read_offset, cast_to=int)
        raise ValueError("Register read offset was not set")


    def solve_arguments(self, fn_addr: Address, args: Sequence[Register]) -> list[ConcreteState]:
        """
        Obtain the concrete values of the `registers` in the function at `fn_address`.
        Uses a "backwards simulation" to find the values of the registers.

        Args:
            fn_address (Address): The address of the function to find the values of the registers
            registers (list[Register]): The registers to find the values of

        Returns:
            list[ConcreteState]: The values of the registers in the function for each path
            to that function
        """
        if not args:
            return [{}]

        simulator = BackwardSimulation(
            project=self.project,
            target=fn_addr,
            cfg=self.cfg,
            variables=list(args),
            max_depth=self.max_depth
        )

        found_states = simulator.simulate()
        return [self._get_args_values(args, state) for state in found_states]

    def _get_args_values(self, args: Sequence[Register], state: angr.SimState) -> ConcreteState:
        concrete_state: ConcreteState = {}
        for arg in args:
            arg.set_ref_states([state])

            if arg.is_concrete():
                concrete_state[arg] = arg.evaluate()[state]
        return concrete_state
